#ifndef HARKIO_HARK_HGTF_M2PG_HPP_
#define HARKIO_HARK_HGTF_M2PG_HPP_

#include <string>
#include <vector>
#include <complex>
#include <harkio2/Version.hpp>
#include <harkio2/Vector3.hpp>
#include <harkio2/defineitem.hpp>
#include <harkio2/hark/DataSet.hpp>
#include <harkio2/InvalidHarkFile.hpp>
#include <harkio2/hark/HarkFileWriter.hpp>

/** @file M2PG.hpp
 * @brief M2PG class and subclasses definitions.
 */

namespace harkio2
{
namespace hark
{
namespace hgtf
{

/** @brief Abstract M2PG data set.
 * This class represents a set of transfer functions
 * from sound sources to microphones.
 */
class M2PG : public harkio2::hark::DataSet
{
public:
    /** @brief Returns the type identifier.
     * @return The type identifier
     */
    virtual std::string GetId() const { return "hgtf::M2PG"; }

    /** @brief Returns the version.
     * Returns the version which this instance actually represents.
     * @return The version number.
     */
    virtual harkio2::Version GetVersion() const = 0;
};

/** @brief M2PG data set version 1.0.
 */
class M2PG1_0 : public M2PG
{
public:
    /** @brief Returns the version.
     * @return The version which this class represents.
     */
    virtual harkio2::Version GetVersion() const { return harkio2::Version(1, 0); }

    /** @brief Transfer function type.
     * This type represents a transfer function.
     */
    typedef std::vector<std::complex<float> > TFType;

    DEFINE_ITEM(nD, unsigned int, ND, unsigned int)
    DEFINE_ITEM(nR, unsigned int, NR, unsigned int)
    DEFINE_ITEM(sourcePositions, std::vector<std::vector<harkio2::Vector3<float> > >,
            SourcePositions, const std::vector<std::vector<harkio2::Vector3<float> > >&)
    DEFINE_ITEM(tfs, std::vector<std::vector<std::vector<TFType> > >,
            TFs, const std::vector<std::vector<std::vector<TFType> > >&)

    /// @brief Implementation of Writable interface.
    void Write(HarkFileWriter& writer) throw (harkio2::InvalidHarkFile)
    {
        writer.Write(*this);
    }

};

}
}
}

#endif // HARKIO_HARK_HGTF_M2PG_HPP_

