
#include <sstream>
#include <cstring>
#include <harkio2/Endian.hpp>
#include <harkio2/hark/HarkFile.hpp>
#include <harkio2/hark/HarkFileBinaryWriter.hpp>
#include <harkio2/hark/DataSet.hpp>
#include <harkio2/hark/hgtf/Fmt.hpp>
#include <harkio2/hark/hgtf/Spec.hpp>
#include <harkio2/hark/hgtf/M2PG.hpp>
#include <harkio2/hark/hgtf/GTF.hpp>
#include <harkio2/hark/hgtf/SM.hpp>

/** @file Helper.cpp
 * @brief implementation of HarkFileBinaryWriter::Helper.
 */

namespace harkio2
{
namespace hark
{

class HarkFileBinaryWriter::Helper
{
    HarkFileBinaryWriter& parent;
public:
    /** @brief Constructs a helper for HarkFileBinaryWriter.
     * @param parent A reference to a parent instance.
     */
    Helper(HarkFileBinaryWriter& parent)
        : parent(parent)
    { }

    /** @brief Writes sizeof(T) bytes to the specified output stream.
     * This function takes account of endians.
     * @param[in] value The value to write.
     * @param[out] out an output stream.
     */
    template <typename T>
    void Write(T value, std::ostream& out)
    {
        value = harkio2::Endian::HostToLittle(value);
        out.write(reinterpret_cast<char*>(&value), sizeof(T));
    }


    /** @brief Writes the specified version.
     * @param v The version.
     */
    void WriteVersion(harkio2::Version v, std::ostream& out)
    {
        Write(static_cast<uint16_t>(v.GetMajor()), out);
        Write(static_cast<uint16_t>(v.GetMinor()), out);
    }

    void WriteHead(const char* id, unsigned int size, std::ostream& out)
    {
        out.write(id, 4);
        Write<uint32_t>(size, out);
    }

    void WriteVector3(const harkio2::Vector3<float>& v, std::ostream& out)
    {
        Write<float>(v.GetX(), out);
        Write<float>(v.GetY(), out);
        Write<float>(v.GetZ(), out);
    }

    void WriteComplex(const std::complex<float>& c, std::ostream& out)
    {
        Write<float>(c.real(), out);
        Write<float>(c.imag(), out);
    }


    void Write(const harkio2::hark::HarkFile& o)
    {
        parent.out->write("HGTF", 4);

        Write(o.GetDataSet("hgtf::Fmt"));

        Write(o.GetDataSet("hgtf::M2PG"));
        Write(o.GetDataSet("hgtf::GTF"));
        Write(o.GetDataSet("hgtf::SM"));

        Write(o.GetDataSet("hgtf::Spec"));
    }

    void Write(boost::shared_ptr<harkio2::hark::DataSet> dataSet)
    {
        if (!dataSet)
        {
            return;
        }
        else
        {
            dataSet->Write(parent);
        }
    }


    void Write(const harkio2::hark::hgtf::Fmt1_0& o)
    {
        std::ostringstream out;
        try
        {
            WriteVersion(o.GetVersion(), out);
            Write<uint32_t>(o.GetFormatId(), out);
            Write<uint32_t>(o.GetNMic(), out);
            Write<uint32_t>(o.GetNFFT(), out);
            Write<uint32_t>(o.GetNImpulseResponse(), out);
            Write<uint32_t>(o.GetFS(), out);

            for (size_t i = 0; i < o.GetMicPositions().size(); ++i)
            {
                const harkio2::Vector3<float>& pos = o.GetMicPositions()[i];
                WriteVector3(pos, out);
            }
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }

        WriteHead("fmt ", out.str().length(), *parent.out);
        *parent.out << out.str();
    }


    void Write(const harkio2::hark::hgtf::Spec1_0& o)
    {
        std::ostringstream out;
        WriteVersion(o.GetVersion(), out);

        char buf[1024];

        try
        {
            memset(buf, 0, 11);
            o.GetDate().copy(buf, 11);
            out.write(buf, 11);

            memset(buf, 0, 128);
            o.GetLocation().copy(buf, 128);
            out.write(buf, 128);

            memset(buf, 0, 128);
            o.GetRobot().copy(buf, 128);
            out.write(buf, 128);

            memset(buf, 0, 128);
            o.GetMicArrayType().copy(buf, 128);
            out.write(buf, 128);

            memset(buf, 0, 128);
            o.GetCreator().copy(buf, 128);
            out.write(buf, 128);

            memset(buf, 0, 1024);
            o.GetMemo().copy(buf, 1024);
            out.write(buf, 1024);
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }

        WriteHead("SPEC", out.str().length(), *parent.out);
        *parent.out << out.str();
    }


    void Write(const harkio2::hark::hgtf::M2PG1_0& o)
    {
        std::ostringstream out;
        try
        {
            WriteVersion(o.GetVersion(), out);
            Write<uint32_t>(o.GetND(), out);
            Write<uint32_t>(o.GetNR(), out);

            for (size_t d = 0; d < o.GetSourcePositions().size(); ++d)
            {
                for (size_t r = 0; r < o.GetSourcePositions()[d].size(); ++r)
                {
                    const harkio2::Vector3<float>& pos = o.GetSourcePositions()[d][r];
                    WriteVector3(pos, out);
                }
            }

            for (size_t d = 0; d < o.GetTFs().size(); ++d)
            {
                for (size_t r = 0; r < o.GetTFs()[d].size(); ++r)
                {
                    for (size_t m = 0; m < o.GetTFs()[d][r].size(); ++m)
                    {
                        for (size_t f = 0; f < o.GetTFs()[d][r][m].size(); ++f)
                        {
                            WriteComplex(o.GetTFs()[d][r][m][f], out);
                        }
                    }
                }
            }
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }

        WriteHead("M2PG", out.str().length(), *parent.out);
        *parent.out << out.str();
    }


    void Write(const harkio2::hark::hgtf::GTF1_0& o)
    {
        std::ostringstream out;
        try
        {
            WriteVersion(o.GetVersion(), out);
            Write<uint32_t>(o.GetNSource(), out);

            for (size_t s = 0; s < o.GetSourcePositions().size(); ++s)
            {
                const harkio2::Vector3<float>& pos = o.GetSourcePositions()[s];
                WriteVector3(pos, out);
            }

            Write<uint32_t>(o.GetNFIndex(), out);

            for (size_t f = 0; f < o.GetFIndexToBinIndexTable().size(); ++f)
            {
                Write<uint32_t>(o.GetFIndexToBinIndexTable()[f], out);
            }

            for (size_t s = 0; s < o.GetTFs().size(); ++s)
            {
                for (size_t m = 0; m < o.GetTFs()[s].size(); ++m)
                {
                    for (size_t f = 0; f < o.GetTFs()[s][m].size(); ++f)
                    {
                        WriteComplex(o.GetTFs()[s][m][f], out);
                    }
                }
            }
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }

        WriteHead("GTF ", out.str().length(), *parent.out);
        *parent.out << out.str();
    }


    void Write(const harkio2::hark::hgtf::SM1_0& o)
    {
        std::ostringstream out;
        try
        {
            WriteVersion(o.GetVersion(), out);
            Write<uint32_t>(o.GetNSource(), out);

            for (size_t s = 0; s < o.GetSourcePositions().size(); ++s)
            {
                const harkio2::Vector3<float>& pos = o.GetSourcePositions()[s];
                WriteVector3(pos, out);
            }

            Write<uint32_t>(o.GetNFIndex(), out);

            for (size_t f = 0; f < o.GetFIndexToBinIndexTable().size(); ++f)
            {
                Write<uint32_t>(o.GetFIndexToBinIndexTable()[f], out);
            }

            for (size_t s = 0; s < o.GetSMs().size(); ++s)
            {
                for (size_t m = 0; m < o.GetSMs()[s].size(); ++m)
                {
                    for (size_t f = 0; f < o.GetSMs()[s][m].size(); ++f)
                    {
                        WriteComplex(o.GetSMs()[s][m][f], out);
                    }
                }
            }
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }

        WriteHead("SM  ", out.str().length(), *parent.out);
        *parent.out << out.str();
    }

};

}
}

