#include <boost/format.hpp>
#include <glibmm/ustring.h>

#include <harkio2/TypedFilename.hpp>
#include <harkio2/harkconfig/Source.hpp>
#include <harkio2/harkconfig/ImpulseResponseListFile.hpp>
#include <harkio2/harkconfig/TSPListFile.hpp>
#include <harkio2/harkconfig/MicArrayLocationFile.hpp>
#include <harkio2/harkconfig/NoiseLocationFile.hpp>

/** @file Helper.cpp
 * @brief implementation of HarkConfigFileXMLWriter::Helper.
 */

using namespace boost;

namespace harkio2
{
namespace harkconfig
{

class HarkConfigFileXMLWriter::Helper
{
    HarkConfigFileXMLWriter& parent;
    private:
        // WriteWithTag only for float&double to fix the format
        void WriteWithTag(const Glib::ustring& tag, const float& value)
        {
            WriteStartTag(tag);
            parent.out << format("%.4f") % value;
            WriteEndTag(tag);
        }
        void WriteWithTag(const Glib::ustring& tag, const double& value)
        {
            WriteStartTag(tag);
            parent.out << format("%.4f") % value;
            WriteEndTag(tag);
        }
public:
    /** @brief Constructs a helper for HarkConfigFileXMLWriter.
     * @param parent A reference to a parent instance.
     */
    Helper(HarkConfigFileXMLWriter& parent)
        : parent(parent)
    { }

    template <typename T>
    void WriteWithTag(const Glib::ustring& tag, const T& value)
    {
        WriteStartTag(tag);
        parent.out << value;
        WriteEndTag(tag);
    }

    template <typename Attr, typename T>
    void WriteWithTag(const Glib::ustring& tag, const Attr& attr, const T& value)
    {
        WriteStartTag(tag, attr);
        parent.out << value;
        WriteEndTag(tag);
    }

    void WriteStartTag(const Glib::ustring& tag)
    {
        parent.out << '<' << tag.raw() << '>';
    }

    void WriteStartTag(const Glib::ustring& tag, const boost::format& attr)
    {
        parent.out << '<' << tag.raw() << ' ' << attr << '>';
    }

    void WriteStartTag(const Glib::ustring& tag, const Glib::ustring& attr)
    {
        parent.out << '<' << tag.raw() << ' ' << attr.raw() << '>';
    }

    void WriteEndTag(const std::string& tag)
    {
        parent.out << "</" << tag << '>';
    }

    template <typename T>
    void WriteVector(const std::vector<T>& data, const Glib::ustring& name)
    {
        WriteStartTag(name, format("size=\"%d\"") % data.size());
        for (size_t i = 0; i < data.size(); ++i)
        {
            Write(data[i]);
        }
        WriteEndTag(name);
    }

    /*
    void Write(const std::vector<harkio2::harkconfig::Source2>& sources)
    {
        WriteStartTag("Sources", format("size=\"%d\"") % sources.size());
        for (size_t i = 0; i < sources.size(); ++i)
        {
            Write(sources[i]);
        }
        WriteEndTag("Sources");
    }
    */

    void Write(const harkio2::harkconfig::Source2& source)
    {
        try
        {
            WriteStartTag("Source");
            Write(source.GetPosition());
            WriteVector(source.GetFiles(), "Files");
            WriteWithTag("CutStart", source.GetCutStart());
            WriteWithTag("CutEnd", source.GetCutEnd());
            if (source.IsCommentInitialized() && source.GetComment().length() > 0)
            {
                WriteWithTag("Comment", source.GetComment().raw());
            }
            WriteEndTag("Source");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    template <typename T>
    void Write(const harkio2::Position<T>& position)
    {
        try
        {
            if (position.IsPosition())
            {
                WriteStartTag("Position", "unit=\"meter\" type=\"cartesian\"");
            }
            else
            {
                WriteStartTag("Position", "unit=\"none\" type=\"cartesian\"");
            }
            WriteWithTag("X", position.GetPositionVector().GetX());
            WriteWithTag("Y", position.GetPositionVector().GetY());
            WriteWithTag("Z", position.GetPositionVector().GetZ());

            WriteEndTag("Position");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    /*
    void Write(const std::vector<harkio2::harkconfig::File1>& files)
    {
        WriteStartTag("Files", format("size=\"%d\"") % files.size());
        for (size_t i = 0; i < files.size(); ++i)
        {
            Write(files[i]);
        }
        WriteEndTag("Files");
    }
    */

    void Write(const harkio2::harkconfig::File1& file)
    {
        try
        {
            WriteStartTag("File");
            WriteWithTag("Name", file.GetName().raw());
            WriteMicIndices(file.GetMicIndices());
            if (file.IsCommentInitialized() && file.GetComment().length() > 0)
            {
                WriteWithTag("Comment", file.GetComment().raw());
            }
            WriteEndTag("File");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    void WriteMicIndices(const std::vector<unsigned int>& indices)
    {
        WriteStartTag("MicIndices", format("size=\"%d\"") % indices.size());
        for (size_t i = 0; i < indices.size(); ++i)
        {
            WriteWithTag("MicIndex", indices[i]);
        }
        WriteEndTag("MicIndices");
    }

    /*
    void Write(const std::vector<harkio2::harkconfig::Mic1>& mics)
    {
        WriteStartTag("Mics", format("size=\"%d\"") % sources.size());
        for (size_t i = 0; i < mics.size(); ++i)
        {
            Write(mics[i]);
        }
        WriteEndTag("Mics");
    }
    */

    void Write(const harkio2::harkconfig::Mic1& mic)
    {
        try
        {
            WriteStartTag("Mic");
            WriteWithTag("MicIndex", mic.GetMicIndex());
            Write(mic.GetPosition());
            if (mic.IsCommentInitialized() && mic.GetComment().length() > 0)
            {
                WriteWithTag("Comment", mic.GetComment().raw());
            }
            WriteEndTag("Mic");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    void Write(const harkio2::harkconfig::NoiseSource1& noise)
    {
        try
        {
            WriteStartTag("NoiseSource");
            WriteWithTag("NoiseIndex", noise.GetNoiseIndex());
            Write(noise.GetPosition());
            if (noise.IsCommentInitialized() && noise.GetComment().length() > 0)
            {
                WriteWithTag("Comment", noise.GetComment().raw());
            }
            WriteEndTag("NoiseSource");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    template <typename T>
    void WriteNumMic(const T& o)
    {
        const char* funcName = "Helper::WriteNumMic";

        try
        {
            if (o.IsNumMicInitialized())
            {
                if (!IsValidNumMic(o.GetNumMic(), o.GetSources()))
                {
                    throw harkio2::InvalidHarkFile((format("%s: invalid NumMic: %d")
                                % funcName % o.GetNumMic()).str());
                }
                WriteWithTag("NumMic", o.GetNumMic());
            }
            else
            {
                if (!IsValidNumMic(o.GetSources()))
                {
                    throw harkio2::InvalidHarkFile((format("%s: the number of MicIndices of Sources is invalid")
                                % funcName).str());
                }
            }
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    std::string ToString(harkio2::FileType::Kind fileType)
    {
        switch (fileType)
        {
        case harkio2::FileType::FLOAT: return "float";
        case harkio2::FileType::WAVE:  return "wave";
        }
    }


    void WriteConfigFile(const harkio2::harkconfig::HarkConfigFile& o)
    {
        const char* funcName = "Helper::Write";

        std::string id = o.GetId();
        size_t filePos = 0;
        if ((filePos = id.find("File")) != std::string::npos)
        {
            id = id.substr(0, filePos);
        }

        try
        {
            WriteStartTag("hark_xml");
            WriteStartTag("hark_config", format("id=\"%s\" major=\"%d\" minor=\"%d\"")
                    % id % o.GetVersion().GetMajor() % o.GetVersion().GetMinor());
            o.Write(parent);
            WriteEndTag("hark_config");
            WriteEndTag("hark_xml");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    void Write(const harkio2::harkconfig::ImpulseResponseListFile2_0& o)
    {
        try
        {
            WriteNumMic(o);
            WriteWithTag("SourceFileType", ToString(o.GetSourceFileType()));
            if (o.IsCommentInitialized() && o.GetComment().length() > 0)
            {
                WriteWithTag("Comment", o.GetComment().raw());
            }
            WriteVector(o.GetSources(), "Sources");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    void Write(const harkio2::harkconfig::TSPListFile2_0& o)
    {
        try
        {
            WriteWithTag("SynchronousAverage", o.GetSynchronousAverage());
            WriteWithTag("OriginalImpulseFile",
                    format("type=\"%s\"") % ToString(o.GetOriginalImpulseFile().GetType()),
                    o.GetOriginalImpulseFile().GetName().raw());
            WriteWithTag("TSPOffset", o.GetTSPOffset());
            WriteWithTag("TSPLength", o.GetTSPLength());
            WriteWithTag("SignalMax", o.GetSignalMax());
            WriteNumMic(o);
            WriteWithTag("SourceFileType", ToString(o.GetSourceFileType()));
            if (o.IsCommentInitialized() && o.GetComment().length() > 0)
            {
                WriteWithTag("Comment", o.GetComment().raw());
            }
            WriteVector(o.GetSources(), "Sources");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    void Write(const harkio2::harkconfig::MicArrayLocationFile2_0& o)
    {
        try
        {
            if (o.IsCommentInitialized() && o.GetComment().length() > 0)
            {
                WriteWithTag("Comment", o.GetComment().raw());
            }
            WriteVector(o.GetMics(), "Mics");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    void Write(const harkio2::harkconfig::NoiseLocationFile2_0& o)
    {
        try
        {
            if (o.IsCommentInitialized() && o.GetComment().length() > 0)
            {
                WriteWithTag("Comment", o.GetComment().raw());
            }
            WriteVector(o.GetNoiseSources(), "NoiseSources");
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

    /** @brief NumMicプロパティと、Sources構造体の整合性を調べる。
     */
    bool IsValidNumMic(unsigned int nmic, const std::vector<harkio2::harkconfig::Source2>& sources)
    {
        unsigned int nmicInFiles;
        try
        {
            for (size_t i = 0; i < sources.size(); ++i)
            {
                nmicInFiles = 0;
                const std::vector<harkio2::harkconfig::File1>& files = sources[i].GetFiles();
                for (size_t j = 0; j < files.size(); ++j)
                {
                    nmicInFiles += files[j].GetMicIndices().size();
                }
                if (nmic != nmicInFiles)
                {
                    return false;
                }
            }
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }

        return true;
    }

    bool IsValidNumMic(const std::vector<harkio2::harkconfig::Source2>& sources)
    {
        if (sources.size() == 0)
        {
            return true;
        }

        try
        {
            unsigned int nmic = sources.at(0).GetFiles().at(0).GetMicIndices().size();
            return IsValidNumMic(nmic, sources);
        }
        catch (const harkio2::NotInitialized& e)
        {
            throw harkio2::InvalidHarkFile(e.what());
        }
    }

};

}
}

