#define BOOST_TEST_MODULE testcase for Position class
#define BOOST_TEST_DYN_LINK
#include "common.hpp"

#include <boost/test/unit_test.hpp>

#include <harkio2/Position.hpp>
#include <stdexcept>
#include <cmath>
#include <iostream>
using namespace harkio2;
using namespace std;

#define RANGE(a, op1, b, op2, c) \
    ((a) op1 (b) && (b) op2 (c))

bool RANGE2(double a, double b, double r)
{
    if (a >= 0)
    {
        return a * (1 - r) <= b && b <= a * (1 + r);
    }
    else
    {
        return a * (1 + r) <= b && b <= a * (1 - r);
    }
}

bool RANGEV(const Vector3<float>& a, const Vector3<float>& b, double r)
{
    return
        RANGE2(a.GetX(), b.GetX(), r) &&
        RANGE2(a.GetY(), b.GetY(), r) &&
        RANGE2(a.GetZ(), b.GetZ(), r);
}

template <typename T>
inline T ToRadian(T deg)
{
    return deg * M_PI / 180;
}

BOOST_AUTO_TEST_CASE(macro_test)
{
    BOOST_CHECK( RANGE(1, <=, 1.1, <=, 1.2) );
    BOOST_CHECK( ! RANGE(1.12, <=, 1.1, <=, 1.2) );

    BOOST_CHECK( RANGE2(2, 2.005, 0.01) );
    BOOST_CHECK( ! RANGE2(2, 2.025, 0.01) );
    BOOST_CHECK( ! RANGE2(0, 0.00001, 0.01) );
}

BOOST_AUTO_TEST_CASE(constructors_test)
{
    Vector3<float> v(1, 0, 0);
    Position<float> p1(v);

    BOOST_CHECK(p1.GetUnitVector() == Vector3<float>(1, 0, 0));
    BOOST_CHECK(p1.GetUnitVector() != Vector3<float>(0, 1, 0));
    BOOST_CHECK_EQUAL(p1.IsPosition(), false);
    BOOST_CHECK_EQUAL(p1.IsDirection(), true);
    BOOST_CHECK_EQUAL(p1.GetAngles().first, 0);
    BOOST_CHECK_EQUAL(p1.GetAngles().second, 0);

    Position<float> p2(v, 5.0f);

    BOOST_CHECK(p2.GetUnitVector() == Vector3<float>(1, 0, 0));
    BOOST_CHECK_EQUAL(p2.IsPosition(), true);
    BOOST_CHECK_EQUAL(p2.IsDirection(), false);
    BOOST_CHECK_EQUAL(p2.GetLength(), 5.0f);
    BOOST_CHECK_EQUAL(p2.GetAngles().first, 0);
    BOOST_CHECK_EQUAL(p2.GetAngles().second, 0);

    Vector3<float> v2(1.0f / sqrt(2.0f), 1.0f / sqrt(2.0f), 0.0f);
    Position<float> p3(v2);

    BOOST_CHECK(RANGE2(1.0f / sqrt(2.0f), p3.GetUnitVector().GetX(), 0.01));
    BOOST_CHECK(RANGE2(1.0f / sqrt(2.0f), p3.GetUnitVector().GetY(), 0.01));
    BOOST_CHECK_EQUAL(p3.GetUnitVector().GetZ(), 0);
    BOOST_CHECK_EQUAL(p3.IsPosition(), false);
    BOOST_CHECK(RANGE2(M_PI / 4, p3.GetAngles().first, 0.01));
    BOOST_CHECK_EQUAL(p3.GetAngles().second, 0);

    Position<float> p4(Vector3<float>(0, 0, 1));

    BOOST_CHECK_EQUAL(p4.GetAngles().first, 0);
    BOOST_CHECK(RANGE2(M_PI / 2, p4.GetAngles().second, 0.01));

    Position<float> p5(Vector3<float>(0, 0, 2));

    BOOST_CHECK_THROW(p5.GetAngles(), std::range_error);
}

BOOST_AUTO_TEST_CASE(factory_test)
{
    Vector3<float> v1(10, 0, 0);
    Position<float> p1 = MakeDirectionFromVector(v1);

    BOOST_CHECK_EQUAL(p1.IsDirection(), true);
    BOOST_CHECK(RANGE2(1.0, p1.GetUnitVector().GetX(), 0.01));
    BOOST_CHECK_EQUAL(p1.GetUnitVector().GetY(), 0);
    BOOST_CHECK_EQUAL(p1.GetUnitVector().GetZ(), 0);
    BOOST_CHECK_EQUAL(p1.GetAngles().first, 0);
    BOOST_CHECK_EQUAL(p1.GetAngles().second, 0);

    Vector3<float> v2(10, 10, 0);
    Position<float> p2 = MakeDirectionFromVector(v2);

    BOOST_CHECK_EQUAL(p2.IsDirection(), true);
    BOOST_CHECK(RANGE2(1.0f / sqrt(2.0f), p2.GetUnitVector().GetX(), 0.01));
    BOOST_CHECK(RANGE2(1.0f / sqrt(2.0f), p2.GetUnitVector().GetY(), 0.01));
    BOOST_CHECK_EQUAL(p2.GetUnitVector().GetZ(), 0);
    BOOST_CHECK(RANGE2(M_PI / 4, p2.GetAngles().first, 0.01));
    BOOST_CHECK_EQUAL(p2.GetAngles().second, 0);

    Position<float> p3 = MakePositionFromVector(Vector3<float>(1, 2, 3));

    BOOST_CHECK_EQUAL(p3.IsPosition(), true);
    BOOST_CHECK(RANGE2(sqrt(14), p3.GetLength(), 0.01));
    BOOST_CHECK(RANGE2(3 / sqrt(14), sin(p3.GetAngles().second), 0.01));
    BOOST_CHECK(RANGE2(2 / sqrt(14), sin(p3.GetAngles().first) * cos(p3.GetAngles().second), 0.01));
    BOOST_CHECK(RANGE2(1 / sqrt(14), cos(p3.GetAngles().first) * cos(p3.GetAngles().second), 0.01));

    Position<float> p4 = MakePositionFromVector<float>(Vector3<float>(1, 2, 3), 10);

    BOOST_CHECK_EQUAL(p4.IsPosition(), true);
    BOOST_CHECK_EQUAL(p4.GetLength(), 10);
    BOOST_CHECK(p3.GetUnitVector() == p4.GetUnitVector());
    BOOST_CHECK(p3.GetAngles().first == p4.GetAngles().first);
    BOOST_CHECK(p3.GetAngles().second == p4.GetAngles().second);

    Position<float> p5 = MakeDirectionFromPolar<float>(M_PI / 4, 0);

    BOOST_CHECK_EQUAL(p5.IsDirection(), true);
    BOOST_CHECK(RANGE2(M_PI / 4, p5.GetAngles().first, 0.01));
    BOOST_CHECK_EQUAL(p5.GetAngles().second, 0);
    BOOST_CHECK(RANGEV(p5.GetUnitVector(), Normalize(Vector3<float>(1, 1, 0)), 0.01));

    float cubeAngle = acos(sqrt(2.0f) / sqrt(3.0f));
    Position<float> p6 = MakeDirectionFromPolar<float>(ToRadian(45.0f), cubeAngle);

    BOOST_CHECK_EQUAL(p6.IsDirection(), true);
    BOOST_CHECK(RANGE2(M_PI / 4, p6.GetAngles().first, 0.01));
    BOOST_CHECK(RANGE2(cubeAngle, p6.GetAngles().second, 0.01));
    BOOST_CHECK(RANGEV(p6.GetUnitVector(), Normalize(Vector3<float>(1, 1, 1)), 0.01));

    Position<float> p7 = MakePositionFromPolar<float>(ToRadian(45.0f), cubeAngle, 2.0);

    BOOST_CHECK_EQUAL(p7.IsPosition(), true);
    BOOST_CHECK(RANGE2(M_PI / 4, p7.GetAngles().first, 0.01));
    BOOST_CHECK(RANGE2(cubeAngle, p7.GetAngles().second, 0.01));
    BOOST_CHECK(RANGEV(p7.GetUnitVector(), Normalize(Vector3<float>(1, 1, 1)), 0.01));
    BOOST_CHECK(RANGE2(2.0, p7.GetLength(), 0.01));

    Position<float> p8 = MakePositionFromPolar<float>(ToRadian(45.0f), M_PI - cubeAngle, 3.0);

    BOOST_CHECK_EQUAL(p8.IsPosition(), true);
    BOOST_CHECK(RANGEV(p8.GetUnitVector(), Normalize(Vector3<float>(-1, -1, 1)), 0.01));
    BOOST_CHECK(RANGE2(ToRadian(180.0f + 45.0f), p8.GetAngles().first, 0.01));
    BOOST_CHECK(RANGE2(cubeAngle, p8.GetAngles().second, 0.01));

    Position<float> p9 = MakePositionFromPolar<float>(ToRadian(45.0f), -(M_PI - cubeAngle), 3.0);

    BOOST_CHECK(RANGEV(p9.GetUnitVector(), Normalize(Vector3<float>(-1, -1, -1)), 0.01));
    BOOST_CHECK(RANGE2(ToRadian(180.0f + 45.0f), p9.GetAngles().first, 0.01));
    BOOST_CHECK(RANGE2(-cubeAngle, p9.GetAngles().second, 0.01));
}

